from typing import List

import matplotlib
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt
from matplotlib.animation import FuncAnimation, PillowWriter
from matplotlib.colors import ListedColormap
from mpl_toolkits.mplot3d.art3d import Poly3DCollection
from scipy.constants import pi
from scipy.spatial import geometric_slerp

from molgri.constants import UNIQUE_TOL, DIM_SQUARE, DIM_LANDSCAPE
from molgri.molecules.parsers import XVGParser, FullGridNameParser, PtParser
from molgri.paths import PATH_OUTPUT_FULL_GRIDS, PATH_INPUT_BASEGRO, PATH_OUTPUT_PT, PATH_INPUT_ENERGIES, \
    PATH_OUTPUT_PLOTS
from molgri.plotting.abstract import Plot3D, AbstractPlot, AbstractMultiPlot
from molgri.plotting.analysis_plots import points_up_to_Ns, test_or_create_Ns
from molgri.space.analysis import vector_within_alpha
from molgri.space.cells import voranoi_surfaces_on_stacked_spheres, voranoi_surfaces_on_sphere
from molgri.space.polytopes import Polytope, IcosahedronPolytope, Cube3DPolytope
from molgri.space.rotobj import build_grid_from_name
from molgri.space.utils import norm_per_axis, normalise_vectors, cart2sphA



class GridPlot(Plot3D):

    def __init__(self, data_name, *, style_type: list = None, plot_type: str = "grid", **kwargs):
        """
        This class is used for plots and animations of grids.

        Args:
            data_name: in the form algorithm_N e.g. randomQ_60
            style_type: a list of style properties like ['empty', 'talk', 'half_dark']
            plot_type: change this if you need unique name for plots with same data_name
            **kwargs:
        """
        if style_type is None:
            style_type = ["talk"]
        super().__init__(data_name, style_type=style_type, plot_type=plot_type, **kwargs)
        self.grid = self._prepare_data()

    def _prepare_data(self) -> np.ndarray:
        my_grid = build_grid_from_name(self.data_name, use_saved=True).get_grid()
        return my_grid

    def _plot_data(self, color="black", s=30, **kwargs):
        self.sc = self.ax.scatter(*self.grid.T, color=color, s=s) #, s=4)

    def create(self, animate_seq=False, **kwargs):
        if "empty" in self.style_type:
            pad_inches = -0.2
        else:
            pad_inches = 0
        x_max_limit = kwargs.pop("x_max_limit", 1)
        y_max_limit = kwargs.pop("y_max_limit", 1)
        z_max_limit = kwargs.pop("z_max_limit", 1)
        super(GridPlot, self).create(equalize=True, x_max_limit=x_max_limit, y_max_limit=y_max_limit,
                                     z_max_limit=z_max_limit, pad_inches=pad_inches, azim=30, elev=10, projection="3d",
                                     **kwargs)
        if animate_seq:
            self.animate_grid_sequence()

    def animate_grid_sequence(self):
        """
        Animate how a grid is constructed - how each individual point is added.

        WARNING - I am not sure that this method always displays correct order/depth coloring - mathplotlib
        is not the most reliable tool for 3d plots and it may change the plotting order for rendering some
        points above others!
        """

        def update(i):
            current_colors = np.concatenate([facecolors_before[:i], all_white[i:]])
            self.sc.set_facecolors(current_colors)
            self.sc.set_edgecolors(current_colors)
            return self.sc,

        facecolors_before = self.sc.get_facecolors()
        shape_colors = facecolors_before.shape
        all_white = np.zeros(shape_colors)

        self.ax.view_init(elev=10, azim=30)
        ani = FuncAnimation(self.fig, func=update, frames=len(facecolors_before), interval=5, repeat=False)
        writergif = PillowWriter(fps=3, bitrate=-1)
        # noinspection PyTypeChecker
        ani.save(f"{self.ani_path}{self.data_name}_{self.plot_type}_ord.gif", writer=writergif, dpi=400)


class PositionGridPlot(GridPlot):

    def __init__(self, data_name, style_type=None, cell_lines=False, plot_type="positions", **kwargs):
        self.cell_lines = cell_lines
        super().__init__(data_name, style_type=style_type, plot_type=plot_type, **kwargs)

    def _prepare_data(self) -> np.ndarray:
        points = np.load(f"{PATH_OUTPUT_FULL_GRIDS}{self.data_name}.npy")
        return points

    def _plot_data(self, color="black", **kwargs):
        points = self._prepare_data()
        points = np.swapaxes(points, 0, 1)
        for i, point_set in enumerate(points):
            self.sc = self.ax.scatter(points[i, :, 0], points[i, :, 1], points[i, :, 2], c=color)
        if self.cell_lines:
            self._draw_voranoi_cells(points)

    def create(self, **kwargs):
        max_norm = np.max(norm_per_axis(self.grid))
        super(PositionGridPlot, self).create(x_max_limit=max_norm, y_max_limit=max_norm, z_max_limit=max_norm, **kwargs)

    def _draw_voranoi_cells(self, points):
        svs = voranoi_surfaces_on_stacked_spheres(points)
        for i, sv in enumerate(svs):
            sv.sort_vertices_of_regions()
            t_vals = np.linspace(0, 1, 2000)
            # plot Voronoi vertices
            self.ax.scatter(sv.vertices[:, 0], sv.vertices[:, 1], sv.vertices[:, 2], c='g')
            # indicate Voronoi regions (as Euclidean polygons)
            for region in sv.regions:
                n = len(region)
                for j in range(n):
                    start = sv.vertices[region][j]
                    end = sv.vertices[region][(j + 1) % n]
                    norm = np.linalg.norm(start)
                    result = geometric_slerp(normalise_vectors(start), normalise_vectors(end), t_vals)
                    self.ax.plot(norm * result[..., 0], norm * result[..., 1], norm * result[..., 2], c='k')


class TrajectoryEnergyPlot(Plot3D):

    def __init__(self, data_name: str, plot_type="trajectory", plot_points=False, plot_surfaces=True,
                 selected_Ns=None, **kwargs):
        self.energies = None
        self.property = "Trajectory positions"
        self.unit = r'$\AA$'
        self.selected_Ns = selected_Ns
        self.N_index = 0
        self.plot_points = plot_points
        self.plot_surfaces = plot_surfaces
        super().__init__(data_name, plot_type=plot_type, **kwargs)

    def add_energy_information(self, path_xvg_file, property_name="Potential"):
        self.plot_type += "_energies"
        file_parser = XVGParser(path_xvg_file)
        self.property, property_index = file_parser.get_column_index_by_name(column_label=property_name)
        self.unit = file_parser.get_y_unit()
        self.energies = file_parser.all_values[:, property_index]

    def _prepare_data(self) -> pd.DataFrame:
        try:
            split_name = self.data_name.split("_")
            path_m1 = f"{PATH_INPUT_BASEGRO}{split_name[0]}"
            path_m2 = f"{PATH_INPUT_BASEGRO}{split_name[1]}"
            try:
                gnp = FullGridNameParser("_".join(split_name[2:]))
                num_b = gnp.get_num_b_rot()
                num_o = gnp.get_num_o_rot()
            except AttributeError:
                print("Warning! Trajectory name not in standard format. Will not be able to perform convergence tests.")
                num_b = 1
                num_o = 1
        except AttributeError:
            raise ValueError("Cannot use the name of the XVG file to find the corresponding trajectory. "
                             "Please rename the XVG file to the same name as the XTC file.")
        path_topology = f"{PATH_OUTPUT_PT}{self.data_name}.gro"
        path_trajectory = f"{PATH_OUTPUT_PT}{self.data_name}.xtc"
        my_parser = PtParser(path_m1, path_m2, path_topology, path_trajectory)
        my_data = []
        for i, molecules in enumerate(my_parser.generate_frame_as_molecule()):
            mol1, mol2 = molecules
            com = mol2.get_center_of_mass()
            try:
                current_E = self.energies[i]
            except TypeError:
                current_E = 0
            my_data.append([*np.round(com), current_E])
        my_df = pd.DataFrame(my_data, columns=["x", "y", "z", f"{self.property} {self.unit}"])
        # if multiple shells present, warn and only use the closest one.
        num_t = len(my_df) // num_b // num_o
        if num_t != 1:
            print("Warning! The pseudotrajectory has multiple shells/translation distances. 2D/3D visualisation is "
                  "most suitable for single-shell visualisations. Only the data from the first shell will be used "
                  "in the visualisation.")
            my_df = get_only_one_translation_distance(my_df, num_t)
        # of all points with the same position, select only the orientation with the smallest energy
        df_extract = groupby_min_body_energy(my_df, target_column=f"{self.property} {self.unit}", N_b=num_b)
        # now potentially reduce the number of orientations tested
        self.selected_Ns = test_or_create_Ns(num_o, self.selected_Ns)
        points_up_to_Ns(df_extract, self.selected_Ns, target_column=f"{self.property} {self.unit}")
        return df_extract

    def _plot_data(self, **kwargs):
        my_df = self._prepare_data()
        # determine min and max of the color dimension
        all_positions = my_df[my_df[f"{self.selected_Ns[self.N_index]}"].notnull()][["x", "y", "z"]].to_numpy()
        all_energies = my_df[my_df[f"{self.selected_Ns[self.N_index]}"].notnull()][f"{self.selected_Ns[self.N_index]}"].to_numpy()
        # sort out what not unique
        _, indices = np.unique(all_positions.round(UNIQUE_TOL), axis=0, return_index=True)
        all_positions = np.array([all_positions[index] for index in sorted(indices)])
        all_energies = np.array([all_energies[index] for index in sorted(indices)])
        # TODO: enable colorbar even if not plotting points
        if self.energies is None:
            self.ax.scatter(*all_positions.T, c="black")
        else:
            if self.plot_surfaces:
                try:
                    self._draw_voranoi_cells(all_positions, all_energies)
                except AssertionError:
                    print("Warning! Sperichal Voranoi cells plot could not be produced. Likely all points are "
                          "not at the same radius. Will create a scatterplot instead.")
                    self.plot_points = True
            cmap = ListedColormap((sns.color_palette("coolwarm", 256).as_hex()))
            im = self.ax.scatter(*all_positions.T, c=all_energies, cmap=cmap)
            #cbar = self.fig.colorbar(im, ax=self.ax)
            #cbar.set_label(f"{self.property} {self.unit}", fontsize=20)
            if not self.plot_points:
                im.set_visible(False)
            self.ax.set_title(r"$N_{rot}$ " + f"= {self.selected_Ns[self.N_index]}", fontsize=30)
            self.ax.view_init(elev=10, azim=60)

    def _draw_voranoi_cells(self, points, colors):
        sv = voranoi_surfaces_on_sphere(points)
        norm = matplotlib.colors.Normalize(vmin=colors.min(), vmax=colors.max())
        cmap = plt.cm.ScalarMappable(cmap="coolwarm", norm=norm)
        cmap.set_array([])
        fcolors = cmap.to_rgba(colors)
        sv.sort_vertices_of_regions()
        for n in range(0, len(sv.regions)):
            region = sv.regions[n]
            polygon = Poly3DCollection([sv.vertices[region]], alpha=1)
            polygon.set_color(fcolors[n])
            self.ax.add_collection3d(polygon)

    def create(self, *args, title=None, **kwargs):
        if title is None:
            title = f"{self.property} {self.unit}"
        super().create(*args, equalize=True, title=title, **kwargs)


class HammerProjectionTrajectory(TrajectoryEnergyPlot, AbstractPlot):

    def __init__(self, data_name: str, plot_type="hammer", figsize=DIM_SQUARE, **kwargs):
        sns.set_context("talk")
        super().__init__(data_name, plot_type=plot_type, plot_surfaces=False, plot_points=True, figsize=figsize,
                         **kwargs)

    def _plot_data(self, **kwargs):

        my_df = self._prepare_data()
        # determine min and max of the color dimension
        # only_index = my_df[my_df[f"{self.selected_Ns[self.N_index]}"].notnull()].index.to_list()
        # all_positions = np.array([*only_index])
        # all_energies = my_df[my_df[f"{self.selected_Ns[self.N_index]}"].notnull()][f"{self.selected_Ns[self.N_index]}"]
        all_positions = my_df[my_df[f"{self.selected_Ns[self.N_index]}"].notnull()][["x", "y", "z"]].to_numpy()
        all_energies = my_df[my_df[f"{self.selected_Ns[self.N_index]}"].notnull()][f"{self.selected_Ns[self.N_index]}"].to_numpy()
        # sort out what not unique
        # _, indices = np.unique(all_positions.round(UNIQUE_TOL), axis=0, return_index=True)
        # all_positions = np.array([all_positions[index] for index in sorted(indices)])
        # all_energies = np.array([all_energies[index] for index in sorted(indices)])
        all_positions = cart2sphA(all_positions)
        x = all_positions[:, 2]
        y = all_positions[:, 1]
        if self.energies is None:
            self.ax.scatter(*all_positions.T, c="black")
        else:
            cmap = ListedColormap((sns.color_palette("coolwarm", 256).as_hex()))
            im = self.ax.scatter(x, y, c=all_energies, cmap=cmap)
            #cbar = self.fig.colorbar(im, ax=self.ax)
            #cbar.set_label(f"{self.property} {self.unit}")
            #self.ax.set_title(r"$N_{rot}$ " + f"= {self.selected_Ns[self.N_index]}", fontsize=40)
        #plt.grid(True)
        self.ax.set_xticklabels([])

    def create(self, *args, **kwargs):
        AbstractPlot.create(self, *args, projection="hammer", **kwargs)


class TrajectoryEnergyMultiPlot(AbstractMultiPlot):

    def __init__(self, list_plots: List[TrajectoryEnergyPlot], figsize=None, **kwargs):
        if figsize is None:
            figsize = (DIM_LANDSCAPE[1]*4, DIM_LANDSCAPE[1])
        super().__init__(list_plots, **kwargs)

    def create(self, *args, **kwargs):
        super().create(*args, projection="3d", **kwargs)


class HammerProjectionMultiPlot(AbstractMultiPlot):

    def __init__(self, list_plots: List[HammerProjectionTrajectory], plot_type="hammer", figsize=None,
                 n_rows=1, n_columns=5, **kwargs):
        if figsize is None:
            figsize = (DIM_LANDSCAPE[1]*n_columns, DIM_LANDSCAPE[1]*n_rows)
        super().__init__(list_plots, plot_type=plot_type, n_rows=n_rows, n_columns=n_columns, figsize=figsize, **kwargs)

    def create(self, *args, projection="hammer", **kwargs):
        super().create(*args, projection=projection, **kwargs)


class GridColoredWithAlphaPlot(GridPlot):
    def __init__(self, data_name, vector: np.ndarray, alpha_set: list, plot_type: str = "colorful_grid", **kwargs):
        super().__init__(data_name, plot_type=plot_type, **kwargs)
        self.alpha_central_vector = vector
        self.alpha_set = alpha_set
        self.alpha_set.sort()
        self.alpha_set.append(pi)

    def _plot_data(self, color=None, **kwargs):
        # plot vector
        self.ax.scatter(*self.alpha_central_vector, marker="x", c="k", s=30)
        # determine color palette
        cp = sns.color_palette("Spectral", n_colors=len(self.alpha_set))
        # sort points which point in which alpha area
        already_plotted = []
        for i, alpha in enumerate(self.alpha_set):
            possible_points = np.array([vec for vec in self.grid if tuple(vec) not in already_plotted])
            within_alpha = vector_within_alpha(self.alpha_central_vector, possible_points, alpha)
            selected_points = [tuple(vec) for i, vec in enumerate(possible_points) if within_alpha[i]]
            array_sel_points = np.array(selected_points)
            self.sc = self.ax.scatter(*array_sel_points.T, color=cp[i], s=30)  # , s=4)
            already_plotted.extend(selected_points)
        self.ax.view_init(elev=10, azim=30)


def create_trajectory_energy_multiplot(data_name, Ns=None, animate_rot=False):
    list_single_plots = []
    max_index = 5 if Ns is None else len(Ns)
    for i in range(max_index):
        tep = TrajectoryEnergyPlot(data_name, plot_points=False, plot_surfaces=True, selected_Ns=Ns, style_type=["talk"])
        tep.N_index = i
        tep.add_energy_information(f"{PATH_INPUT_ENERGIES}{data_name}")
        list_single_plots.append(tep)
    TrajectoryEnergyMultiPlot(list_single_plots, n_columns=max_index, n_rows=1).create_and_save(animate_rot=animate_rot,
                                                                                                elev=10, azim=60)


def create_hammer_multiplot(data_name, Ns=None):
    list_single_plots = []
    max_index = 5 if Ns is None else len(Ns)
    for i in range(max_index):
        tep = HammerProjectionTrajectory(data_name, selected_Ns=Ns, style_type=["talk"])
        tep.N_index = i
        tep.add_energy_information(f"{PATH_INPUT_ENERGIES}{data_name}")
        list_single_plots.append(tep)
    HammerProjectionMultiPlot(list_single_plots, n_columns=max_index, n_rows=1).create_and_save()


class PolytopePlot(Plot3D):

    def __init__(self, data_name: str, num_divisions=3, faces=None, projection=False, **kwargs):
        """
        Plotting (some faces of) polyhedra, demonstrating the subdivision of faces with points.

        Args:
            data_name:
            num_divisions: how many levels of faces subdivisions should be drawn
            faces: a set of indices indicating which faces to draw
            projection: if True display points projected on a sphere, not on faces
            **kwargs:
        """
        self.num_divisions = num_divisions
        self.faces = faces
        self.projection = projection
        plot_type = f"polytope_{num_divisions}"
        super().__init__(data_name, fig_path=PATH_OUTPUT_PLOTS, plot_type=plot_type, style_type=["empty"],
                         **kwargs)

    def _prepare_data(self) -> Polytope:
        if self.data_name == "ico":
            ico = IcosahedronPolytope()
        else:
            ico = Cube3DPolytope()
        for n in range(self.num_divisions):
            ico.divide_edges()
        return ico

    def _plot_data(self, **kwargs):
        if self.faces is None and self.data_name == "ico":
            self.faces = {12}
        elif self.faces is None:
            self.faces = {3}
        ico = self._prepare_data()
        ico.plot_points(self.ax, select_faces=self.faces, projection=self.projection)
        ico.plot_edges(self.ax, select_faces=self.faces)


def groupby_min_body_energy(df: pd.DataFrame, target_column: str, N_b: int) -> pd.DataFrame:
    """
    Take a dataframe with positions and energies and return only one row per COM position of the second molecule,
    namely the one with lowest energy.

    Args:
        df: dataframe resulting from a Pseudotrajectory with typical looping over:
            rotations of origin
                rotations of body
                    translations must be already filtered out by this point!
        target_column: name of the column in which energy values are found
        N_b: number of rotations around body for this PT.

    Returns:
        a DataFrame with a number of rows equal original_num_of_rows // N_b
    """
    # in case that translations have been removed, the index needs to be re-set
    df.reset_index(inplace=True, drop=True)
    start_len = len(df)
    new_df = df.loc[df.groupby(df.index // N_b)[target_column].idxmin()]
    assert len(new_df) == start_len // N_b
    return new_df


def get_only_one_translation_distance(df: pd.DataFrame, N_t: int, distance_index=0) -> pd.DataFrame:
    start_len = len(df)
    assert distance_index < N_t, f"Only {N_t} different distances available, " \
                                 f"cannot get the distance with index {distance_index}"
    new_df = df.iloc[range(0, len(df), N_t)]
    assert len(new_df) == start_len // N_t
    return new_df


from typing import Tuple

import numpy as np
import pandas as pd
from numpy.typing import NDArray
from scipy.constants import pi

from molgri.space.utils import angle_between_vectors, random_sphere_points, random_quaternions, \
    standardise_quaternion_set, randomise_quaternion_set_signs


def vector_within_alpha(central_vec: NDArray, side_vector: NDArray, alpha: float) -> bool or NDArray:
    """
    Answers the question: which angles between both vectors or sets of vectors are within angle angle alpha?

    Args:
        central_vec: a single vector of shape (d,) or an array of vectors of shape (n1, d)
        side_vector: a single vector of shape (d,) or an array of vectors of shape (n2, d)
        alpha: in radians, angle around central_vec where we check if side_vectors occur

    Returns:
        array of bools of shape (n1, n2) if both vectors are 2D
                       of shape (n2,) if central_vec is 1D and side_vec 2D
                       of shape (n1,) if central_vec is 2D and side_vec 1D
                       of shape (1,) if both vectors are 1D
    """
    return angle_between_vectors(central_vec, side_vector) < alpha


def count_points_within_alpha(array_points: np.ndarray, central_vec: np.ndarray, alpha: float):
    # since a list of True or False is returned, the sum gives the total number of True elements
    return np.sum(vector_within_alpha(central_vec, array_points, alpha))


def random_axes_count_points(array_points: np.ndarray, alpha: float, num_random_points: int = 1000):
    central_vectors = random_sphere_points(num_random_points)
    all_ratios = np.zeros(num_random_points)
    for i, central_vector in enumerate(central_vectors):
        num_within = count_points_within_alpha(array_points, central_vector, alpha)
        all_ratios[i] = num_within/len(array_points)
    return all_ratios


def random_quaternions_count_points(array_points: np.ndarray, alpha: float, num_random_points: int = 1000):
    central_vectors = random_quaternions(num_random_points)
    # # use half-sphere in both examples
    central_vectors = randomise_quaternion_set_signs(central_vectors)
    array_points = randomise_quaternion_set_signs(array_points)
    # for i, line in enumerate(central_vectors):
    #     if line[0] < 0:
    #         central_vectors[i, :] = -central_vectors[i, :]
    # for i, line in enumerate(array_points):
    #     if line[0] < 0:
    #         array_points[i, :] = -array_points[i, :]
    all_ratios = np.zeros(num_random_points)
    for i, central_vector in enumerate(central_vectors):
        num_within = count_points_within_alpha(array_points, central_vector, alpha)
        all_ratios[i] = num_within/len(array_points)
    return all_ratios


if __name__ == "__main__":
    from molgri.space.rotobj import build_rotations
    from molgri.constants import GRID_ALGORITHMS

    N = 40

    for algo in GRID_ALGORITHMS[:-1]:
        quats = build_rotations(N, algo=algo).rotations.as_quat()
        ratios = random_quaternions_count_points(quats, alpha=pi/6)
        print(algo, np.average(ratios), np.std(ratios))


def prepare_statistics(point_array: NDArray, alphas: list, d=3, num_rand_points=1000) -> Tuple:
    """
    Prepare statistics data that is used for ViolinPlots, either based on 3D points or on quaternions.

    Args:
        point_array: an array of points for which coverage is tested, must be of the shape (N, d)
        alphas: a list of angles in radians at which convergence is tested
        d: number of dimensions (3 or 4)
        num_rand_points: number of random axes around which convergence is tested

    Returns:
        (data_frame_summary, data_frame_full)
    """
    if alphas is None:
        alphas = [pi / 6, 2 * pi / 6, 3 * pi / 6, 4 * pi / 6, 5 * pi / 6]
    assert d == 3 or d == 4, f"Statistics available only for 3 or 4 dimensions, not d={d}"
    assert point_array.shape[1] == d, f"Provided array not in shape (N, {d})"
    if d == 3:
        sphere_surface = 4 * pi
        cone_area_f = lambda a: 2 * pi * (1 - np.cos(a))
        actual_coverages_f = random_axes_count_points
    else:
        # explanation: see https://scialert.net/fulltext/?doi=ajms.2011.66.70&org=11
        sphere_surface = 2 * pi ** 2  # full 4D sphere has area 2pi^2 r^3
        cone_area_f = lambda a: 1 / 2 * sphere_surface * (2 * a - np.sin(2 * a)) / np.pi
        actual_coverages_f = random_quaternions_count_points
    # write out short version ("N points", "min", "max", "average", "SD"
    columns = ["alphas", "ideal coverages", "min coverage", "avg coverage", "max coverage", "standard deviation"]
    ratios_columns = ["coverages", "alphas", "ideal coverage"]
    ratios = [[], [], []]
    data = np.zeros((len(alphas), 6))  # 5 data columns for: alpha, ideal coverage, min, max, average, sd
    for i, alpha in enumerate(alphas):
        cone_area = cone_area_f(alpha) # cone area of 4D cone is
        ideal_coverage = cone_area / sphere_surface
        actual_coverages = actual_coverages_f(point_array, alpha, num_random_points=num_rand_points)
        ratios[0].extend(actual_coverages)
        ratios[1].extend([alpha] * num_rand_points)
        ratios[2].extend([ideal_coverage] * num_rand_points)
        data[i][0] = alpha
        data[i][1] = ideal_coverage
        data[i][2] = np.min(actual_coverages)
        data[i][3] = np.average(actual_coverages)
        data[i][4] = np.max(actual_coverages)
        data[i][5] = np.std(actual_coverages)
    alpha_df = pd.DataFrame(data=data, columns=columns)
    alpha_df = alpha_df.set_index("alphas")
    ratios_df = pd.DataFrame(data=np.array(ratios).T, columns=ratios_columns)
    return alpha_df, ratios_df


def write_statistics(stat_data: pd.DataFrame, full_data: pd.DataFrame,
                     path_summary: str, path_full_data: str,
                     num_random, name, dimensions, print_message: bool = False):
    # first message (what measure you are using)
    newline = "\n"
    if dimensions == 3:
        m1 = f"STATISTICS: Testing the coverage of grid {name} using {num_random} " \
             f"random points on a sphere. This statistics is useful for the position grid."
    elif dimensions == 4:
        m1 = f"STATISTICS: Testing the coverage of quaternions {name} using {num_random} " \
             f"random quaternions. This statistics is useful for the relative orientations."
    else:
        raise ValueError("Dimensions must be 3 or 4.")
    m2 = f"We select {num_random} random axes and count the number of grid points that fall within the angle" \
         f"alpha (selected from [pi / 6, 2 * pi / 6, 3 * pi / 6, 4 * pi / 6, 5 * pi / 6]) of this axis. For an" \
         f"ideally uniform grid, we expect the ratio of num_within_alpha/total_num_points to equal the ratio" \
         f"area_of_alpha_spherical_cap/area_of_sphere, which we call ideal coverage."
    if print_message:
        print(m1)
        print(stat_data)
    # dealing with the file
    with open(path_summary, "w") as f:
        f.writelines([m1, newline, newline, m2, newline, newline])
    stat_data.to_csv(path_summary, mode="a")
    full_data.to_csv(path_full_data, mode="w")